PFX_TOKEN_SHORT_VALIDITY = {'hours': 12}
PFX_TOKEN_LONG_VALIDITY = {'days': 30}

PFX_COOKIE_DOMAIN = None
PFX_COOKIE_SECURE = True
PFX_COOKIE_SAMESITE = 'None'

PFX_OPENAPI_PATH = "doc/api/"
PFX_OPENAPI_FILENAME = "openapi"
PFX_OPENAPI_TEMPLATE = {}
